# -*-python-*-

__copyright__ = "Copyright (c) 1999-2014 CollabNet, Inc. All rights reserved."

# NOTE: Reliable file update semantics employed herein adopted from
# http://blog.gocept.com/2013/07/15/reliable-file-updates-with-python/

import sys
import os
import SourceForge
import ctfauthorizer
import tempfile
import time
from svn import repos, fs, core


PBP_CACHE_LIFETIME_SECS = 60 * 5
ENABLE_LOGGING = False


class TFSvnPBPCache:
  def __init__(self, system_id, rootpath, rootname, username):
    self.pbpCacheDir = os.path.join(rootpath, '.teamforge', 'pbpcache')
    self.pbpCacheFile = os.path.join(self.pbpCacheDir, username)

  def _stringify(self, elm):
    if isinstance(elm, dict):
      return dict(map(lambda i: (self._stringify(i[0]), self._stringify(i[1])),
                      elm.items()))
    elif isinstance(elm, list):
      return map(lambda x: self._stringify(x), elm)
    elif isinstance(elm, unicode):
      return elm.encode('utf-8')
    else:
      return elm

  def get(self):
    try:
      # json module is new in Python 2.6
      import json
      if os.path.exists(self.pbpCacheFile):
        pbpCache = json.load(open(self.pbpCacheFile, 'r'))
        if pbpCache['expires'] >= int(time.time()):
          return self._stringify(pbpCache['pbps'])
    except:
      pass
    return None

  def set(self, pbps):
    try:
      # json module is new in Python 2.6
      import json

      def _update_cache_file(pbps):
        # Create a sibling temporary file with the new content, then
        # rename the temporary file atop the original, and fsync() the
        # parent directory.
        tmp_fileno, tmp_filename = tempfile.mkstemp('w', dir=self.pbpCacheDir)
        tmp_fp = os.fdopen(tmp_fileno, 'w')
        json.dump({ 'expires': int(time.time() + PBP_CACHE_LIFETIME_SECS),
                    'pbps': pbps,
                  }, tmp_fp)
        tmp_fp.close()
        os.rename(tmp_filename, self.pbpCacheFile)
        dir_fd = os.open(self.pbpCacheDir, os.O_DIRECTORY)
        os.fsync(dir_fd)
        os.close(dir_fd)

      # Create missing cache parent directories.
      if not os.path.exists(self.pbpCacheDir):
        os.makedirs(self.pbpCacheDir)

      # If the cache file is missing, create it.  Otherwise, hold a lock on
      # it (OS-allowing) and then use the write-replace update method.
      #
      # Adapted (for Python 2.4 compatibility) from LockedOpen as presented at
      # http://blog.gocept.com/2013/07/15/reliable-file-updates-with-python/
      if not os.path.exists(self.pbpCacheFile):
        _update_cache_file(pbps)
      else:
        fp = None
        try:
          try:
            import fcntl
            fp = open(self.pbpCacheFile, 'r+')
            while True:
              fcntl.flock(fp, fcntl.LOCK_EX)
              fp_new = open(self.pbpCacheFile, 'r+')
              if os.path.sameopenfile(fp.fileno(), fp_new.fileno()):
                fp_new.close()
                break
              else:
                fp.close()
                fp = fp_new
            _update_cache_file(pbps)
          except ImportError:
            _update_cache_file(pbps)
        finally:
          if fp:
            fp.close()
    except:
      pass


class DebugLogger:
  def __init__(self, rootpath, username):
    # If ENABLE_LOGGING isn't set or we fail to open the log file,
    # this will remain None and turn off logging.
    self.log_file_fp = None

    if ENABLE_LOGGING:
      # Determine the static part of the logged messages.  The uuid
      # module is new in Python 2.5, so we'll include a per-instance
      # UUID in those lines only if we can generate one.
      username = username or '-'
      try:
        import uuid
        self.prefix = "%s [%s]" % (uuid.uuid1().hex, username)
      except:
        self.prefix = "[%s]" % (username)

      # Try to open the logfile.  If this fails, we won't be logging
      # anything.
      try:
        log_dir = os.path.join(rootpath, '.teamforge')
        if not os.path.exists(log_dir):
          os.makedirs(log_dir)
        self.log_file_fp = open(os.path.join(log_dir, 'log'), 'a')
      except:
        pass

  def write(self, s):
    if self.log_file_fp:
      self.log_file_fp.write("%s %s %s\n" %
                             (time.asctime(), self.prefix, s))
      self.log_file_fp.flush()


DIRECTORY_ACCESS_DEEPDENIAL = 0 # Denied for this directory and all children
DIRECTORY_ACCESS_SHALLOW = 1    # That's (sh)allow (depth 1 access granted)


_access_values = {
  -1: 'UNKNOWN',
  0: 'NONE',
  1: 'VIEW',
  2: 'COMMIT',
  }


class TFSvnAuthorizer:
  # Create an interface for checking path read access to the
  # Subversion repository named ROOTNAME, located at the absolute path
  # ROOTPATH and associated with TeamForge SYSTEM_ID.  Perform access
  # checks as USERNAME.  If READABLE_PARENTDIRS is set, grant access
  # to otherwise unreadable parent directories of readable paths.
  def __init__(self, rootname, rootpath, username, system_id,
               readable_parentdirs=True):
    # The basics: who wants to see stuff in what repositories.
    self.rootname = rootname
    self.rootpath = rootpath
    self.username = username
    self.system_id = system_id
    self.readable_parentdirs = readable_parentdirs
    self.logger = DebugLogger(rootpath, username)

    # Cache of directory access results:  DIRECTORY -> DIRECTORY_ACCESS_*
    # All directories are in one of three possible states:
    #
    #    1. access unknown (not present in the dict)
    #
    #    2. access granted, either directly or indirectly (in the dict
    #       with DIRECTORY_ACCESS_SHALLOW)
    #
    #    3. access denied (in the dict with DIRECTORY_ACCESS_DEEPDENIAL)
    self.perm_cache = { }

    # See if there's a valid, local cache of the role paths we can use.
    pbpCache = TFSvnPBPCache(system_id, rootpath, rootname, username)
    self.logger.write("Fetching PBPs from disk cache...")
    raw_pbps = pbpCache.get()

    # If we didn't get the raw PBPs, we'll have to fetch them via SOAP.
    if raw_pbps is not None:
      self.logger.write("Success")
    else:
      self.logger.write("Failure")
      scm_service = SourceForge.getSOAPClient().service
      key = SourceForge.createScmRequestKey()

      # If asked to use the old getRolePaths() API, do so.  Otherwise,
      # try to use the getRolePathsList() API and fall back to
      # getRolePaths() only if the former fails.
      if not SourceForge.getBoolean("svn.authnz.use_role_paths_api", "false"):
        try:
          raw_pbps = self._get_role_paths_list(scm_service, key)
        except:
          self.logger.write("Failed; falling back to older API")
          raw_pbps = self._get_role_paths(scm_service, key)
      else:
        raw_pbps = self._get_role_paths(scm_service, key)

      # Try to cache what we found.
      pbpCache.set(raw_pbps)

    # Register the PBPs.
    self.logger.write("Instantiating CtfAuthorizer...")
    self.ctf_auth = ctfauthorizer.CtfAuthorizer(raw_pbps)
    self.logger.write("CtfAuthorizer instance created")

    self.logger.write("Initialized CtfAuthorizer with PBPs from %d roles" % (len(raw_pbps)))

  def _get_role_paths_list(self, scm_service, key):
    self.logger.write("Fetching PBPs via SOAP (getRolePathsList)...")
    raw_pbps_suds = scm_service.getRolePathsList(key, self.username,
                                                 self.system_id, self.rootname)
    self.logger.write("Success")
    self.logger.write("Parsing PBPs...")
    raw_pbps = {}
    for item in raw_pbps_suds:
      if item is not None:
        raw_pbps[str(item.repoRoleId)] = []
        for i in item.repoPathAndPerm:
          raw_pbps[str(item.repoRoleId)].append(str(i))
    self.logger.write("Completed")
    return raw_pbps

  def _get_role_paths(self, scm_service, key):
    self.logger.write("Fetching PBPs via SOAP (getRolePaths)...")
    raw_pbps_suds = scm_service.getRolePaths(key, self.username,
                                             self.system_id, self.rootname)
    self.logger.write("Success")
    self.logger.write("Parsing PBPs...")
    raw_pbps = []
    for item in raw_pbps_suds:
      raw_pbps.append(str(item))
    self.logger.write("Completed")
    return raw_pbps

  def determine_global_access(self):
    self.logger.write("Determining global access...")
    value = self.ctf_auth.determine_global_access()
    self.logger.write("Global access determined: %s"
                      % (_access_values.get(value, '** UNDEFINED **')))
    return value

  def _check_path_read_access(self, path):
    self.logger.write("Checking access for path '%s'..." % (path))
    value = self.ctf_auth.has_non_recursive_access(path, 1)
    self.logger.write("Path access determined: %s"
                      % (_access_values.get(value, '** UNDEFINED **')))
    return value

  def _scan_dir(self, revision_root, path, deadline):
    """Recursively scan a repository directory to determine which
    paths are accessible."""

    # If we've already answered the auth question for this directory,
    # return it.
    if self.perm_cache.has_key(path):
      return self.perm_cache[path]

    # Initialize the access granting state.
    grant_access = DIRECTORY_ACCESS_DEEPDENIAL

    # If the directory matches, grant access.  Otherwise, we have to
    # see if we earn access indirectly.
    if self._check_path_read_access(path):
      grant_access = DIRECTORY_ACCESS_SHALLOW
      return grant_access

    # If we're not allowing deep access checks, this is as far as
    # we'll get.
    if not self.readable_parentdirs:
      return grant_access

    # If atleast one children at any depth level of this directory
    # is readable, we have to grant access to this directory.
    # So read the directory, tracking sets of files and subdirs.
    files = []
    subdirs = []

    # Read the directory entries.
    try:
      entries = fs.dir_entries(revision_root, path)
    except core.SubversionException, e:
      if e.apr_err == core.SVN_ERR_FS_NOT_FOUND:
        # this must have been a path deleted from rev; don't try to recurse.
        entries = {}

    for entry in entries.values():
      if entry.kind == core.svn_node_dir:
        subdirs.append(entry)
      elif entry.kind == core.svn_node_file:
        files.append(entry)

    # If we don't have access yet, check out files.  If any file
    # grants access, also grant access to this directory.
    if grant_access != DIRECTORY_ACCESS_SHALLOW:
      for file in files:
        if (deadline is not None) and (time.time() > deadline):
          raise TFSvnAuthorizerTimeout("Timed out scanning directory files.")
        full_name = os.path.join(path, file.name)
        if self._check_path_read_access(full_name):
          grant_access = DIRECTORY_ACCESS_SHALLOW
          break

    # If we don't have access yet, recurse on subdirs.  If any subdir
    # grants access, also grant access to this directory.
    if grant_access != DIRECTORY_ACCESS_SHALLOW:
      for subdir in subdirs:
        if (deadline is not None) and (time.time() > deadline):
          raise TFSvnAuthorizerTimeout("Timed out scanning directory subdirectories.")
        if self._scan_dir(revision_root, os.path.join(path, subdir.name), deadline):
          grant_access = DIRECTORY_ACCESS_SHALLOW
          break
        else:
          self.perm_cache[path + '/' + subdir.name] = \
            DIRECTORY_ACCESS_DEEPDENIAL

    # If access has been granted for this directory, cache that fact.
    if grant_access == DIRECTORY_ACCESS_SHALLOW:
      self.perm_cache[path] = grant_access

    # Return the grant.
    return grant_access

  def check_path_access(self, path, rev, is_file, timeout=None):
    # Checking file access is straightforward, so do so.
    if is_file:
      if self._check_path_read_access(path):
        return 1
      else:
        return 0

    # If we have a cached permission for this (directory) path, return it.
    if self.perm_cache.has_key(path):
      if self.perm_cache[path] == DIRECTORY_ACCESS_SHALLOW:
        return 1
      elif self.perm_cache[path] == DIRECTORY_ACCESS_DEEPDENIAL:
        return 0
    else:
      # Recursively scan a repository directory to determine which
      # paths are accessible.
      repository = repos.open(self.rootpath)
      fs_ptr = repos.fs(repository)
      deadline = (timeout is not None) and (time.time() + timeout) or None
      try:
        self.perm_cache[path] = self._scan_dir(fs.revision_root(fs_ptr, rev),
                                               path, deadline)
      except TFSvnAuthorizerTimeout, e:
        self.logger.write(str(e))
        raise

    access = self.perm_cache[path]
    if access == DIRECTORY_ACCESS_SHALLOW:
      return 1
    elif access == DIRECTORY_ACCESS_DEEPDENIAL:
      return 0

    # At this point, we've should have a result.  But just in case,
    # default to not-authorized.
    return self.perm_cache.get(path, 0)


class TFSvnAuthorizerTimeout(Exception): pass
