__copyright__ = "Copyright (c) 2016 CollabNet, Inc. All rights reserved."

#---------------------------------------------------------------------------

# Target size for a given chunk of transmitted indexed data.
INDEX_CHUNK_THRESHOLD_SIZE = 1024 * 1024

# Maximum allowed size of an indexed file's contents.
MAX_INDEXED_FILE_SIZE = 1024 * 1024 * 1

# Maximum file size for which we will attempt character encoding detection.
MAX_ENCODING_DETECTION_FIZE_SIZE = 1024 * 100

# svn:mime-type patterns we will attempt to index.
MIME_TYPE_WHITELIST = [
    'text/*',
    '*/xml',
    '*/json',
    '*/*+xml',
    '*/*+json',
    ]

# svn:mime-type patterns we will not index (unless they also match the
# whitelist).
MIME_TYPE_BLACKLIST = [
    'application/octet-stream',
    'application/pdf',
    'application/zip',
    'application/jar',
    'image/*',
    ]

#---------------------------------------------------------------------------

import sys
import os
import urllib
import json
import fnmatch
import hashlib
try:
    from cStringIO import StringIO
except:
    from StringIO import StringIO
import svn.repos, svn.fs, svn.core, svn.delta
import SourceForge
import requests
import warnings
from urllib3 import exceptions
warnings.simplefilter('ignore', exceptions.HTTPWarning)


# ----------------------------------------------------------------------
# TeamForge Indexing API Abstraction
# ----------------------------------------------------------------------

class _object:
    def __init__(self, **kwargs):
        self.__dict__.update(kwargs)


class TeamForgeSearchableRepository:
    def __init__(self, system_id, repos_data={}):
        self.system_id = system_id
        self.project_id = repos_data['projectId']
        self.id = repos_data['id']
        self.path = repos_data['repositoryDirectory']
        self.name = os.path.basename(self.path)
        self.last_indexed_revision = None
        self.last_indexed_paths = []
        self.base_commit_id = repos_data.get('lastIndexedCommits', {}).get('default')
        if self.base_commit_id is not None:
            try:
                self.last_indexed_revision = int(self.base_commit_id)
                self.last_indexed_paths = []
            except:
                try:
                    last_indexed = json.loads(self.base_commit_id)
                    last_rev = last_indexed.get('revision', 0)
                    self.last_indexed_revision = int(last_rev)
                    self.last_indexed_paths = last_indexed.get('paths', [])
                except:
                    pass
        self.last_modified_date = repos_data['lastModifiedDate']
        params_dict = {}
        for param in repos_data.get('params', []):
            for key, value in param.items():
                if not params_dict.has_key(key):
                    params_dict[key] = []
                params_dict[key].append(value)
        self.params = params_dict

    def __repr__(self):
        return ('TeamForgeSearchableRepository systemId(%s) id(%s) '
                'projectId(%s) path(%s)' % (self.system_id,
                                            self.id,
                                            self.project_id,
                                            self.path))


class TeamForgeIndexerAPIError(Exception):
    """An error occured in the TeamForge Indexer API."""
    pass


class TeamForgeIndexerAPINotAvailableError(TeamForgeIndexerAPIError):
    """The TeamForge Indexer API is not available."""
    pass


class TeamForgeIndexerAPI:
    """Interface for the TeamForge Code Search and Indexing API."""

    def __init__(self):
        self.api_base_url = SourceForge.getAPIServerBaseUrl() + \
                            '/ctfrest/codesearchindex/v1'

    def get_searchable_repositories(self, system_id):
        url = self.api_base_url + '/' + system_id + '/repositories'
        data = {
            'requestKey': SourceForge.createScmRequestKey(),
            }
        response = requests.post(url, json=data, verify=False)
        self._raise_for_response_status(response)
        response_data = json.loads(response.text)
        repositories = {}
        for repos_data in response_data:
            try:
                repository = TeamForgeSearchableRepository(system_id,
                                                           repos_data)
                repositories[repository.name] = repository
            except:
                pass
        return repositories

    def file_id_hash(self, s):
        """Convert id into a sha1 hash"""
        hash = hashlib.new('sha1')
        hash.update(s)
        return hash.hexdigest()

    def get_bulk_action_data_chunk(self, repository, path, action,
                                   size=None, content=None):
        """Return a string suitable for use as the entirety of, or just a
        chunk of, the ElasticSearch/TeamForge bulk indexing API input."""
        assert action in ('index', 'delete')
        fp = StringIO()
        file_id = '|'.join([repository.system_id, repository.name,
                            'default', path])
        if len(file_id) > 512:
            file_id = self.file_id_hash(file_id)
        if action == 'index':
            fname, ext = os.path.splitext(path)
            if len(ext) > 0 and ext[0] == '.':
                ext = ext[1:]
            action = {
                'index': {
                    '_type': 'file',
                    '_id': file_id,
                    }
                }
            if callable(content):
                content = content(path)
            fields = {
                'path': path,
                'ref': 'default',
                'repository': '|'.join([repository.system_id, repository.id]),
                'repositoryId': repository.id,
                'projectId': repository.project_id,
                'extension': ext,
                'size': size,
                'content': content,
                }
            return "%s\n%s\n" % (json.dumps(action), json.dumps(fields))
        else:
            action = {
                'delete': {
                    '_type': 'file',
                    '_id': file_id,
                    }
                }
            return "%s\n" % (json.dumps(action))

    def send_index_actions(self, repository, bulk_action_data):
        """Update the index with a REVISION number and ElasticSearch bulk
        indexing API data BULK_ACTION_DATA."""
        url = self.api_base_url + '/index/' + \
              '/'.join([repository.system_id, repository.name, 'default'])
        data = {
            'requestKey': SourceForge.createScmRequestKey(),
            'actions': bulk_action_data,
            }
        if repository.base_commit_id:
            data['baseCommitId'] = repository.base_commit_id
        response = requests.post(url, json=data, verify=False)
        self._raise_for_response_status(response)

    def finalize(self, repository, indexed_paths, revision):
        url = self.api_base_url + '/index/' + \
              '/'.join([repository.system_id, repository.name, 'default'])
        data = {
            'requestKey': SourceForge.createScmRequestKey(),
            'commitId': self._get_commit_id(revision, indexed_paths),
            }
        if repository.base_commit_id:
            data['baseCommitId'] = repository.base_commit_id
        response = requests.post(url, json=data, verify=False)
        self._raise_for_response_status(response)

    def _raise_for_response_status(self, response):
        if response.status_code == 409:
            raise TeamForgeIndexerAPIError("Operation conflicted with other "
                                           "indexing operations")
        if response.status_code == 501:
            raise TeamForgeIndexerAPINotAvailableError("Code search is not "
                                                       "configured on the "
                                                       "server")
        try:
            response.raise_for_status()
        except Exception, e:
            raise TeamForgeIndexerAPIError(str(e))

    def _get_commit_id(self, revision, indexed_paths):
        indexed_paths.sort(svn.core.svn_path_compare_paths)
        return json.dumps({
            'revision': str(revision),
            'paths': indexed_paths,
            }, sort_keys=True)


# ----------------------------------------------------------------------
# Indexing Logger (to be subclassed)
# ----------------------------------------------------------------------

class SvnIndexerLogger:
    def __init__(self, repository):
        """Initialize a logger instance for a given
        TeamForgeSearchableRepository instance."""
        pass

    def indexing_skipped_uptodate(self, revision):
        """Called to report that there's nothing new to index is
        already up-to-date."""
        raise NotImplementedError()

    def indexing_started(self, old_revision, new_revision):
        """Called when indexing begins."""
        raise NotImplementedError()

    def dir_index_started(self, path):
        """Called when indexing begins for subtree."""
        raise NotImplementedError()

    def dir_index_completed(self, path):
        """Called when indexing is completed for subtree."""
        raise NotImplementedError()

    def file_index_added(self, path):
        """Called when a file's content is added to the index."""
        raise NotImplementedError()

    def file_index_deleted(self, path):
        """Called when a file's content is removed from the index."""
        raise NotImplementedError()

    def file_index_skipped_too_large(self, path, size):
        """Called when a file's content is too large to index."""
        raise NotImplementedError()

    def file_index_skipped_encoding(self, path):
        """Called when a file's content encoding is un-indexable."""
        raise NotImplementedError()

    def file_index_skipped_mimetype(self, path):
        """Called when a file's MIME type is blacklisted."""
        raise NotImplementedError()

    def indexing_finalized(self, revision):
        """Called when the indexing batch is completed."""
        raise NotImplementedError()


# ----------------------------------------------------------------------
# Indexer
# ----------------------------------------------------------------------


class SvnIndexerBlacklistedMimeTypeError(Exception):
    pass


class SvnIndexer:
    def __init__(self, api, repository):
        self.api = api
        self.repository = repository
        self.old_revision = None
        self.logger = None
        self.action_buffer = None

    def set_logger(self, logger):
        self.logger = logger

    def begin(self, old_revision, new_revision):
        self.old_revision = old_revision
        self._log_begin(old_revision, new_revision)

    def dir_index_started(self, path):
        self._log_dir_index_started(path)

    def dir_index_completed(self, path):
        self._log_dir_index_completed(path)

    def last_indexed_rev(self):
        return self.repository.last_indexed_revision

    def last_indexed_paths(self):
        return self.repository.last_indexed_paths

    def index(self, path, size, utf8_content_func):
        if size > MAX_INDEXED_FILE_SIZE:
            if self.logger:
                self.logger.file_index_skipped_too_large(path, size)
            return
        content = None
        try:
            content = utf8_content_func(path)
        except SvnIndexerBlacklistedMimeTypeError, e:
            if self.logger:
                self.logger.file_index_skipped_mimetype(path)
        except:
            if self.logger:
                self.logger.file_index_skipped_encoding(path)
        if content is None:
            return
        bulk_action_data = self.api.get_bulk_action_data_chunk(self.repository,
                                                               path, 'index',
                                                               size, content)
        self._queue_or_dispatch_bulk_action(bulk_action_data)
        self._log_index(path)

    def delete(self, path):
        bulk_action_data = self.api.get_bulk_action_data_chunk(self.repository,
                                                               path, 'delete')
        self._queue_or_dispatch_bulk_action(bulk_action_data)
        self._log_delete(path)

    def finalize(self, indexed_paths, revision):
        if self.action_buffer is not None:
            self.api.send_index_actions(self.repository, self.action_buffer)
            self.action_buffer = None
        self.api.finalize(self.repository, indexed_paths, revision)
        self._log_finalize(revision)

    def _queue_or_dispatch_bulk_action(self, bulk_action_data):
        if len(bulk_action_data) > INDEX_CHUNK_THRESHOLD_SIZE:
            self.api.send_index_actions(self.repository, bulk_action_data)
        else:
            if self.action_buffer is None:
                self.action_buffer = bulk_action_data
            else:
                self.action_buffer += bulk_action_data
            if len(self.action_buffer) > INDEX_CHUNK_THRESHOLD_SIZE:
                self.api.send_index_actions(self.repository, self.action_buffer)
                self.action_buffer = None

    def _log_begin(self, old_revision, new_revision):
        if self.logger:
            self.logger.indexing_started(old_revision, new_revision)

    def _log_dir_index_started(self, path):
        if self.logger:
            self.logger.dir_index_started(path)

    def _log_dir_index_completed(self, path):
        if self.logger:
            self.logger.dir_index_completed(path)

    def _log_finalize(self, revision):
        if self.logger:
            self.logger.indexing_finalized(revision)

    def _log_index(self, path):
        if self.logger:
            self.logger.file_index_added(path)

    def _log_delete(self, path):
        if self.logger:
            self.logger.file_index_deleted(path)


# ----------------------------------------------------------------------
# Subversion Delta Editor Indexer Driver
# ----------------------------------------------------------------------

def indexer_delete_dir(indexer, rev_root, path):
    for name, dirent in svn.fs.dir_entries(rev_root, path).items():
        fullpath = (path != '/' and path or '') + '/' + name
        if dirent.kind == svn.core.svn_node_dir:
            indexer_delete_dir(indexer, rev_root, fullpath)
        elif dirent.kind == svn.core.svn_node_file:
            indexer.delete(fullpath)


class SvnRevisionRangeDeltaEditor(svn.delta.Editor):
    def __init__(self, indexer, base_path, base_rev_root, rev_root):
        self.indexer = indexer
        self.base_rev_root = base_rev_root
        self.rev_root = rev_root
        self.base_path = base_path
        self.revision = svn.fs.revision_root_revision(self.rev_root)

    def delete_entry(self, path, revision, parent_baton, pool=None):
        path = self._full_path(path)
        kind = svn.fs.check_path(self.base_rev_root, path)
        if kind == svn.core.svn_node_dir:
            indexer_delete_dir(self.indexer, self.base_rev_root, path)
        elif kind == svn.core.svn_node_file:
            self.indexer.delete(path)

    def add_file(self, path, parent_baton,
                             copyfrom_path, copyfrom_revision, file_pool=None):
        return self._full_path(path)

    def open_file(self, path, parent_baton, base_revision, file_pool=None):
        return self._full_path(path)

    def close_file(self, file_baton, text_checksum, pool=None):
        self._index_file(file_baton)

    def _full_path(self, path):
        return (self.base_path != '/' and self.base_path or '') + '/' + path

    def _get_utf8_file_content(self, path):
        mime_type, encoding = self._get_mime_data(path)

        # If there's a MIME type, see if it's one that we'll try to
        # index (explicitly whitelisted or at least not blacklisted).
        if mime_type is not None:
            whitelisted = False
            for pattern in MIME_TYPE_WHITELIST:
                if fnmatch.fnmatch(mime_type, pattern):
                    whitelisted = True
            if not whitelisted:
                for pattern in MIME_TYPE_BLACKLIST:
                    if fnmatch.fnmatch(mime_type, pattern):
                        raise SvnIndexerBlacklistedMimeTypeError(mime_type)

        # Get the file's contents.
        stream = svn.fs.file_contents(self.rev_root, path)
        fp = StringIO()
        while 1:
            chunk = svn.core.svn_stream_read(stream,
                                             svn.core.SVN_STREAM_CHUNK_SIZE)
            if not chunk:
                break
            fp.write(chunk)
        svn.core.svn_stream_close(stream)
        filecontent = fp.getvalue()
        fp.close()

        # If there's no encoding specified at the file is lower than
        # our encoding detection size limit, we'll try to detect the
        # encoding.  Otherwise, we assume it's UTF-8 and pray for the
        # best.
        if encoding is None:
            size = svn.fs.file_length(self.rev_root, path)
            if size <= MAX_ENCODING_DETECTION_FIZE_SIZE:
                encoding = SourceForge.detect_encoding(filecontent, 0.5)
            else:
                encoding = 'utf-8'
        return SourceForge.transcode_to_utf8(filecontent, encoding)

    def _get_mime_data(self, path):
        mime_type = None
        encoding = None
        proplist = svn.fs.node_proplist(self.rev_root, path)
        svn_mime_type = proplist.get(svn.core.SVN_PROP_MIME_TYPE)
        if svn_mime_type:
            mime_parts = map(lambda x: x.strip(), svn_mime_type.split(';'))
            type_subtype = mime_parts[0].lower()
            parameters = {}
            for part in mime_parts[1:]:
                name, value = part.split('=', 1)
                parameters[name] = value
            mime_type = type_subtype
            encoding = parameters.get('charset')
        return mime_type, encoding

    def _index_file(self, path):
        size = svn.fs.file_length(self.rev_root, path)
        self.indexer.index(path, size, self._get_utf8_file_content)


def index_repository(repository, indexer, index_rev=None):
    """Index a TeamForgeSearchableRepository."""

    repos_path = svn.core.svn_path_canonicalize(repository.path)
    repos_obj = svn.repos.open(repos_path)
    fs_obj = svn.repos.fs(repos_obj)
    youngest_rev = svn.fs.youngest_rev(fs_obj)
    assert(index_rev is None or index_rev <= youngest_rev)
    if index_rev is None:
        index_rev = youngest_rev
    last_indexed_rev = indexer.last_indexed_rev()
    if last_indexed_rev is None:
        last_indexed_rev = 0

    def _authz_callback(fsroot, path, pool):
        return 1

    def _subsumes(path1, path2):
        if path1 == path2:
            return True
        if ((len(path1) < len(path2)) and
            (path1 + '/' == path2[:len(path1) + 1])):
            return True
        return False

    # Which paths should we index?
    if repository.params.has_key('indexedPaths'):
        paths_to_index = repository.params['indexedPaths']
    else:
        paths_to_index = ['/trunk']

    # Reduce index paths, eliminate overlapped ones.
    index_dirs = []
    paths_to_index = map(lambda x: svn.core.svn_path_canonicalize(
                             x[0] == '/' and x or '/' + x),
                         paths_to_index)
    paths_to_index.sort(svn.core.svn_path_compare_paths)
    for path in paths_to_index:
        try:
            for index_dir in index_dirs:
                assert not _subsumes(index_dir, path)
            index_dirs.append(path)
        except AssertionError:
            pass

    # We need to see if the set of paths-to-be-indexed has changed
    # since our last indexing run.
    old_indexed_paths = set(indexer.last_indexed_paths())
    new_indexed_paths = set(index_dirs)
    added_paths = list(new_indexed_paths - old_indexed_paths)
    removed_paths = list(old_indexed_paths - new_indexed_paths)

    # If we've already indexed to/past our requested index_rev and the
    # configured paths haven't changed, there's nothing more to do.
    if ((last_indexed_rev >= index_rev) and
        not added_paths and not removed_paths):
        if indexer.logger:
            indexer.logger.indexing_skipped_uptodate(last_indexed_rev)
        return

    rev_root = svn.fs.revision_root(fs_obj, index_rev)
    base_rev_root = svn.fs.revision_root(fs_obj, last_indexed_rev)
    zero_rev_root = svn.fs.revision_root(fs_obj, 0)

    indexer.begin(last_indexed_rev, index_rev)

    # Remove paths we no longer want to be indexed.
    for index_dir in removed_paths:
        base_rev_kind = svn.fs.check_path(base_rev_root, index_dir)
        if base_rev_kind == svn.core.svn_node_dir:
            indexer.dir_index_started(index_dir)
            indexer_delete_dir(indexer, base_rev_root, index_dir)
            indexer.dir_index_completed(index_dir)

    # Update paths we do want to be indexed.
    for index_dir in index_dirs:
        rev_kind = svn.fs.check_path(rev_root, index_dir)
        base_rev_kind = svn.fs.check_path(base_rev_root, index_dir)

        if rev_kind == svn.core.svn_node_dir:
            if (base_rev_kind == svn.core.svn_node_dir and
                index_dir not in added_paths):
                # Path is a directory in both the old and new
                # revisions and has been indexed before.
                old_root = base_rev_root
                old_path = index_dir
            else:
                # Path is a newly appearing directory -- either newly
                # appearing as a directory in the repository, or newly
                # appearing in the list of indexed paths.
                old_root = zero_rev_root
                old_path = '/'
            indexer.dir_index_started(index_dir)
            e = SvnRevisionRangeDeltaEditor(indexer, index_dir,
                                            old_root, rev_root)
            editor, baton = svn.delta.make_editor(e)
            svn.repos.dir_delta2(old_root, old_path, '',
                                 rev_root, index_dir, editor, baton,
                                 _authz_callback, 0,
                                 svn.core.svn_depth_infinity, False, True)
            indexer.dir_index_completed(index_dir)
        elif base_rev_kind == svn.core.svn_node_dir:
            # Path was a directory, but is no longer.  Remove it.
            indexer.dir_index_started(index_dir)
            indexer_delete_dir(indexer, base_rev_root, index_dir)
            indexer.dir_index_completed(index_dir)
    indexer.finalize(paths_to_index, index_rev)
