#!/usr/bin/env python2

#
#
# SourceForge(r) Enterprise Edition
# Copyright 2007-2014 CollabNet, Inc.  All rights reserved.
# http://www.collab.net
#

"""Sourceforge post-commit script - given a repository and a revision number, create the commit
   in SourceForge and add the files to it."""

import re, sys, os
import svn
import SourceForge
import SubversionUtil

class Error(Exception): pass

def doIt(pool, repository, revision, systemId):
    # global log
    hostName = SourceForge.getRequired('sfmain.integration.listener_host')
    hostPort = SourceForge.getRequired('sfmain.integration.listener_port')
    if systemId == None:
        systemId = SourceForge.getRequired('external_system_id')
    try:
        maxFiles = int(SourceForge.getRequired('sfmain.integration.maxFilesListedInTFCommitObject'))
    except:
        maxFiles = 250

    svnlook = SubversionUtil.createSVNLook(repository, rev = revision)
    changes = svnlook.changed()
    changes = SubversionUtil.generalizeChanges(changes)

    # Make a list of only paths, from a list of (op, prop, path) tuples
    changedPaths = []
    versions = []
    statuses = []
    refFilenames = []
    refVersions = []
    statusMap = {'D': 'Deleted', 'A': 'Added', 'M': 'Moved', 'C': 'Copied', 'U': 'Modified'}

    cachesToInvalidate = []
    vmToInvalidate = []

    for changeRec in changes:
        changedPaths.append(unicode(changeRec[2], 'utf-8'))
        versions.append(str(revision))

        # In the branding repository, see if this commit changed i18n
        # `.properties' or `.vm' template files.  If so, we need to
        # invalidate the associated caches.
        if repository == "/sf-svnroot/branding":
            if changeRec[2].endswith(".properties") and changeRec[2].find('i18n/') != -1:
                fileName = changeRec[2]
                fileName = fileName[fileName.index('i18n/')+5:fileName.index('.properties')]
                fileName = fileName.replace('/','.')
                cachesToInvalidate.append(fileName)

            if changeRec[2].endswith(".vm") and changeRec[2].find('templates/mail/') != -1:
                fileName = changeRec[2]
                fileName = fileName[fileName.index('templates/mail/')+15:fileName.index('.vm')]
                fileName = fileName.replace('/','.')
                vmToInvalidate.append(fileName)

        # Append the change status, favoring the content status over
        # the property status.
        if changeRec[0]:
            statuses.append(statusMap.get(changeRec[0], ''))
        elif changeRec[1]:
            statuses.append(statusMap.get(changeRec[1], ''))
        else:
            statuses.append('')

        # Append any copyfrom information that might be present.
        if len(changeRec) > 3 and changeRec[3] != None:
            refFilenames.append(unicode(changeRec[3], 'utf-8'))
            refVersions.append(str(changeRec[4]))
        else:
            refFilenames.append(unicode('', 'utf-8'))
            refVersions.append('')

    author = svnlook.author()
    logmsg = svnlook.log()
    diff = svnlook.diff(max_length=25001)

    # Now hook up the SOAP call to create the commit

    log.write("Instantiating the scm listener")
    scm = SourceForge.getSOAPClient()
    commitId = None
    try:
        log.write("Making createCommit() call")
        key = SourceForge.createScmRequestKey()

        try:
            diff = re.sub(r'[\x00-\x08\x0B-\x0C\x0E-\x1F\x7f-\xff]', "", diff)
            diff = unicode(diff, 'utf-8')
        except Exception, e:
            log.write('Unable to encode diff: ' + str(e) + '\n')
            diff = 'Unable to generate diff.  (Diff content is not decodable.)'

        # If there are more changes that is allowed by configuration,
        # we'll slice the various lists, then add one more faux entry
        # to each to describe the fact that we've cropped/dropped
        # data.
        numFiles = len(changedPaths)
        if maxFiles > 0 and numFiles > maxFiles:
            numDropped = numFiles - maxFiles
            changedPaths = changedPaths[:maxFiles]
            changedPaths.append('')
            versions = versions[:maxFiles]
            versions.append(str(revision))
            refVersions = refVersions[:maxFiles]
            refVersions.append('')
            refFilenames = refFilenames[:maxFiles]
            refFilenames.append('')
            statuses = statuses[:maxFiles]
            if numDropped > 999:
                statuses.append('COMMIT(999+ files not visible)')
            else:
                statuses.append('COMMIT(%d files not visible)' % (numDropped))

        commitId = scm.service.createCommit(key,
                                            unicode(author, 'utf-8'),
                                            systemId,
                                            repository,
                                            unicode(logmsg, 'utf-8'),
                                            changedPaths,
                                            versions,
                                            statuses,
                                            refFilenames,
                                            refVersions,
                                            diff)
    except Exception, inst:
        log.write('Failed to create commit: ' + str(inst) + '\n')
        import traceback
        traceback.print_tb(sys.exc_info()[2], None, None)
        return 1

    log.write('createCommit returned the commit id:' + commitId)

    # prepare and make the call to add files to the commit

    # Invalidate the cache if any commits need a cache refresh
    if len(cachesToInvalidate) != 0:
        cacheInvalidator = SourceForge.getSOAPClient("SourceForge", "50")
        for cache in cachesToInvalidate:
            cacheInvalidator.service.invalidateResourceBundleCache(cache)

    if len(vmToInvalidate) != 0:
        cacheInvalidator = SourceForge.getSOAPClient("SourceForge", "50")
        for cache in vmToInvalidate:
            cacheInvalidator.service.invalidateEmailTemplate(cache)
    # If we were using a trace file, close the file now that we are done writing
    # log.close()

    return 0

def main():
    args = sys.argv
    if len(args) < 3:
        print 'Error: not enough args: ', args
        return 1

    repository = unicode(args[1], 'utf-8')
    revision = int(args[2])

    if len(args) >= 4:
        systemId = args[3]
    else:
        systemId = None

    if not SubversionUtil.isSVNRepository(repository):
        raise Error, "path '%s' doesn't look like a repository" % repository

    print 'repository: ', repository
    print 'revision: ', revision
    svn.core.run_app(doIt, repository, revision, systemId)


if __name__ == "__main__":
    log = SourceForge.getLogFile('post-commit.log')
    log.setLogging(False)

    try:
        main()
    except Exception, inst:
        log.write('Failed in post-commit: ' + str(inst) + '\n')
        import traceback
        traceback.print_tb(sys.exc_info()[2], None, log)
        log.close()
        raise
