# -*-python-*-
import difflib
import itertools
import json

def _opcode_text(opcode):
  """Return a descriptive string for the diff operation code OPCODE."""
  return ({
    ' ': 'context',
    '+': 'add',
    '-': 'remove',
    }).get(opcode, 'unknown')


def _opcode_is_context(opcode):
  """Return whether diff operation code OPCODE describes context."""
  return opcode == ' '


def _context_line(line_num_left, line_num_right, text):
  """Return a JSON object representing a context line."""
  return {
    'kind': 'context',
    'lineNumLeft': line_num_left,
    'lineNumRight': line_num_right,
    'text': text,
    }


def _udiff_line(opcode, line_num_left, line_num_right, text):
  """Return a JSON object representing a unified diff line."""
  return {
    'kind': 'udiff-%s' % (opcode == '-' and 'remove' or 'add'),
    'lineNumLeft': opcode != '+' and line_num_left or None,
    'lineNumRight': opcode != '-' and line_num_right or None,
    'text': text,
    }


def _sdiff_line(line_num_left, text_left,
                line_num_right, text_right, is_context):
  """Return a JSON object representing a sidebyside diff line."""
  if not is_context:
    line_num_left = text_left is not None and line_num_left or None
    line_num_right = text_right is not None and line_num_right or None
  return {
    'kind': 'sdiff',
    'lineNumLeft': line_num_left,
    'textLeft': text_left,
    'lineNumRight': line_num_right,
    'textRight': text_right,
    }


def _get_idiff_tokens(text_line, hint_line):
  """Return a set of JSON tokens describing the intraline changes in
  TEXT_LINE as described by HINT_LINE."""
  tokens = []
  in_context = None
  hint_line = hint_line.rstrip()
  hints = itertools.groupby(hint_line, lambda x: x == ' ' and x or '*')
  last_index = 0

  def _add_token(kind, text):
    unicode(text, 'utf-8') # Ensure we didn't split a multi-byte character
    tokens.append({
      'kind': kind,
      'text': text,
      })

  for key, group in hints:
    group_len = len(list(group))
    _add_token(key == ' ' and 'match' or 'change',
               text_line[last_index:last_index+group_len])
    last_index += group_len
  if len(text_line) > len(hint_line):
    _add_token('match', text_line[len(hint_line):])
  return tokens


def _context_gap(line_num_left_first, line_num_left_last,
                 line_num_right_first, line_num_right_last):
  return {
    'kind': 'gap',
    'lineNumLeftFirst': line_num_left_first,
    'lineNumLeftLast': line_num_left_last,
    'lineNumRightFirst': line_num_right_first,
    'lineNumRightLast': line_num_right_last,
    }


def _calculate_gap(context_start, context_end, max_context, is_end):
  """Given the first and last 0-based line numbers of context and the
  maximum allowed context, return a 2-tuple of the first and last
  lines of context which should be replaced with a context gap
  indicator."""

  num_context_lines = context_end + 1 - context_start

  # prelude context
  if context_start == 0:
    if num_context_lines > max_context:
      return context_start, context_end - max_context
  # postlude context
  elif is_end:
    if num_context_lines > max_context:
      return context_start + max_context, context_end
  # intradiff context
  else:
    if num_context_lines > (max_context * 2):
      return context_start + max_context, context_end - max_context
  return None, None


def _cmp_udiff_lines(a, b):
  """Compare unified diff JSON objects for the purposes of sorting."""
  if a['kind'] == b['kind']:
    if a['kind'] == 'udiff-remove':
      return cmp(a['lineNumLeft'], b['lineNumLeft'])
    else:
      return cmp(a['lineNumRight'], b['lineNumRight'])
  if a['kind'] == 'udiff-remove':
    return -1
  else:
    return 1


class JsonDiff:
  """A class which encapsulates the generation of difference information
  between two ordered sequences of strings (typically the individual
  lines of two text files)."""

  def __init__(self, leftlines, rightlines, num_context_lines=3):
    self.num_context_lines = num_context_lines
    self.diff = difflib.Differ().compare(leftlines, rightlines)

  def simple(self):
    state = {
      'linecount1': 0,
      'linecount2': 0,
      }
    def _diffline_to_dict(line, state):
      opcode = line[0]
      # For now, ignore intraline diff hints.
      if opcode == '?': return None
      if opcode == ' ':
        state['linecount1'] += 1
        state['linecount2'] += 1
      elif opcode == '-':
        state['linecount1'] += 1
      elif opcode == '+':
        state['linecount2'] += 1
      return _udiff_line(opcode, state['linecount1'],
                         state['linecount2'], line[2:])
    lines = filter(None, map(lambda x: _diffline_to_dict(x, state),
                             list(self.diff)))
    return {
      'diffLines': lines,
      'diffStyle': 'simple',
      }

  def unified(self):
    linecount1 = 0
    linecount2 = 0
    lines = []
    diff_lines = []
    first_context_lineno = None
    has_diff = False

    for line in list(self.diff):
      opcode = line[0]
      assert opcode in (' ', '-', '+', '?')

      if opcode == '?':
        # intraline diff hint
        if not diff_lines:
          continue
        last_line = diff_lines[-1]
        try:
          last_line_idiff = _get_idiff_tokens(last_line['text'], line[2:])
          last_line['idiff'] = last_line_idiff
          del(last_line['text'])
        except:
          pass
      elif _opcode_is_context(opcode):
        # context
        linecount1 += 1
        linecount2 += 1
        if diff_lines:
          diff_lines.sort(_cmp_udiff_lines)
          lines.extend(diff_lines)
          diff_lines = []
        lines.append(_context_line(linecount1, linecount2, line[2:]))
        if first_context_lineno is None:
          first_context_lineno = len(lines)
      else:
        # difference
        if (self.num_context_lines and first_context_lineno is not None):
          last_context_lineno = len(lines)
          gap_first, gap_last = _calculate_gap(first_context_lineno - 1,
                                               last_context_lineno - 1,
                                               self.num_context_lines, False)
          if gap_first is not None:
            lines[gap_first:gap_last + 1] = [
              _context_gap(lines[gap_first]['lineNumLeft'],
                           lines[gap_last]['lineNumLeft'],
                           lines[gap_first]['lineNumRight'],
                           lines[gap_last]['lineNumRight'])
              ]
        if opcode == '-':
          linecount1 += 1
        else:
          linecount2 += 1
        first_context_lineno = last_context_lineno = None
        diff_lines.append(_udiff_line(opcode, linecount1,
                                      linecount2, line[2:]))
        has_diff = True
    if diff_lines:
      diff_lines.sort(_cmp_udiff_lines)
      lines.extend(diff_lines)
    elif self.num_context_lines:
      last_context_lineno = len(lines) - 1
      gap_first, gap_last = _calculate_gap(first_context_lineno - 1,
                                           last_context_lineno - 1,
                                           self.num_context_lines, True)
      if gap_first is not None:
        lines[gap_first:] = [
          _context_gap(lines[gap_first]['lineNumLeft'],
                       lines[gap_last]['lineNumLeft'],
                       lines[gap_first]['lineNumRight'],
                       lines[gap_last]['lineNumRight'])
          ]
    if not has_diff:
      lines = []
    return {
      'diffLines': lines,
      'diffStyle': 'unified',
      }

  def sidebyside(self):
    linecount1 = 0
    linecount2 = 0
    lines = []
    first_context_lineno = None
    has_diff = False

    for line in list(self.diff):
      opcode = line[0]
      assert opcode in (' ', '-', '+', '?')

      if opcode == '?':
        last_line = lines[-1]
        if last_line['textRight'] is not None:
          try:
            last_line['idiffRight'] = _get_idiff_tokens(last_line['textRight'],
                                                        line[2:])
            del(last_line['textRight'])
          except:
            pass
        else:
          try:
            last_line['idiffLeft'] = _get_idiff_tokens(last_line['textLeft'],
                                                       line[2:])
            del(last_line['textLeft'])
          except:
            pass
      elif _opcode_is_context(opcode):
        linecount1 += 1
        linecount2 += 1
        lines.append(_context_line(linecount1, linecount2, line[2:]))
        if first_context_lineno is None:
          first_context_lineno = len(lines)
      else:
        has_diff = True
        if (self.num_context_lines and first_context_lineno is not None):
          last_context_lineno = len(lines)
          gap_first, gap_last = _calculate_gap(first_context_lineno - 1,
                                               last_context_lineno - 1,
                                               self.num_context_lines, False)
          if gap_first is not None:
            lines[gap_first:gap_last + 1] = [
              _context_gap(lines[gap_first]['lineNumLeft'],
                           lines[gap_last]['lineNumLeft'],
                           lines[gap_first]['lineNumRight'],
                           lines[gap_last]['lineNumRight'])
              ]
        first_context_lineno = last_context_lineno = None
        if opcode == '-':
          linecount1 += 1
          lines.append(_sdiff_line(linecount1, line[2:],
                                   linecount2, None, False))
        else:
          linecount2 += 1
          if (lines
              and not lines[-1]['kind'] == 'context'
              and lines[-1].get('textRight') is None
              and lines[-1].get('idiffRight') is None):
            lines[-1]['lineNumRight'] = linecount2
            lines[-1]['textRight'] = line[2:]
          else:
            lines.append(_sdiff_line(linecount1, None,
                                     linecount2, line[2:], False))
    if (self.num_context_lines and first_context_lineno is not None):
      last_context_lineno = len(lines) - 1
      gap_first, gap_last = _calculate_gap(first_context_lineno - 1,
                                           last_context_lineno - 1,
                                           self.num_context_lines, True)
      if gap_first is not None:
        lines[gap_first:] = [
          _context_gap(lines[gap_first]['lineNumLeft'],
                       lines[gap_last]['lineNumLeft'],
                       lines[gap_first]['lineNumRight'],
                       lines[gap_last]['lineNumRight'])
          ]
    if not has_diff:
      lines = []
    return {
      'diffLines': lines,
      'diffStyle': 'sidebyside',
      }

