# -*-python-*-
#
# This file implements a custom ViewVC authorization module which
# authenticates user access based on TeamForge path-based
# permissions.

__copyright__ = "Copyright (c) 1999-2014 CollabNet, Inc. All rights reserved."

import sys
import os
import string
import vcauth
import vclib
import SourceForge
from ctfsvnauthz import TFSvnAuthorizer, TFSvnAuthorizerTimeout


class ViewVCAuthorizer(vcauth.GenericViewVCAuthorizer):
  """Custom ViewVC authorization module for TeamForge."""
  
  def __init__(self, root_lookup_func, username=None, params={}):
    """Create a GenericViewVCAuthorizer object which will be used to
    validate that USERNAME has the permissions needed to view version
    control repositories (in whole or in part).  PARAMS is a
    dictionary of custom parameters for the authorizer."""

    self.query_dict = params['query_dict'] # The query parameters
    self.rootname = self.query_dict['root']
    self.roottype, self.rootpath = root_lookup_func(self.rootname)

    # A single-item array, where the item is a universal access
    # determination (1=yes, 0=no, None=dunno); or None if we haven't
    # queried for universal access yet.
    self.global_access = None

    # Get some configuration values from the SourceForge module.
    # These were designed for the SVN REST API, but we'll use them,
    # too.
    self.readable_parentdirs = \
      SourceForge.getBoolean("svn.api.readable_parentdirs", "true")
    self.path_authz_timeout = \
      float(SourceForge.get("svn.api.path_authz_timeout", "10"))

    # Setup (for Subversion) a TFSvnAuthorizer that will handle our
    # access checks.
    self.svn_authorizor = None
    if self.roottype == "svn":
      username = username or 'nobody'
      system_id = self.query_dict['system']
      self.svn_authorizor = TFSvnAuthorizer(self.rootname, self.rootpath,
                                            username, system_id,
                                            self.readable_parentdirs)
                                            
  def check_root_access(self, rootname):
    """Return 1 if the associated username is permitted to read ROOTNAME."""

    # Root access is always allowed.
    return 1

  def check_universal_access(self, rootname):
    """Return 1 if the associated username is permitted to read ALL
    paths in ROOTNAME, 0 if username is permitted to read NO paths in
    ROOTNAME, or None if no such universal access determination can be
    made."""

    # If we already know the answer, return it.
    if self.global_access is not None:
      return self.global_access[0]
    
    # Check (and remember) the global view access access.  For CVS,
    # global view access is always granted.  For Subversion, we ask
    # our TFSvnAuthorizer.
    if self.roottype == "cvs":
      self.global_access = [1]
    else:
      pbp_access = self.svn_authorizor.determine_global_access()
      self.global_access = [pbp_access and 1 or None]

    # Return the access determination.
    return self.global_access[0]

  def check_path_access(self, rootname, path_parts, pathtype, rev=None):
    """Return 1 if the associated username is permitted to read
    revision REV of the path PATH_PARTS (of type PATHTYPE) in
    repository ROOTNAME."""

    # Short circuit the authorization check when the user is known to have
    # a universal access determination (either granting or denying access).
    # (This should never be "None" for CVS repositories.)
    univ_access = self.check_universal_access(rootname)
    if univ_access is not None:
      return univ_access

    # If we get here, we should be looking at an SVN repository.  Ask
    # the TFSvnAuthorizer about path access.
    path = '/%s' % (string.join(path_parts, '/'))
    try:
      return self.svn_authorizor.check_path_access(path, rev,
                                                   pathtype == vclib.FILE,
                                                   self.path_authz_timeout)
    except TFSvnAuthorizerTimeout:
      return 0
