__copyright__ = "Copyright (c) 2014 CollabNet, Inc. All rights reserved."


import os
import re
import logging
import urllib2
import time

from app.scripts import BaseInitializationScript
import app.util
import service.util
from suds.client import Client
from suds.cache import ObjectCache

class InitializationScript(BaseInitializationScript):

    scriptDescription = "Register SCM integration"

    def __getExternalSystemId(self, basePath):
        externalSystemId = None

        try:
            with open(os.path.join(basePath, '.scm.properties')) as scmProperties:
                for line in scmProperties:
                    key, val = line.partition("=")[::2]
                    if key == 'external_system_id':
                        externalSystemId = val.strip()
                        break
        except:
            pass
        return externalSystemId

    def __portSegment(self, scheme, port):
        if scheme == "https":
            if port == 443:
                return ""
            else:
                return ":%d" % port
        else:
            if port == 80:
                return ""
            else:
                return ":%d" % port

    def __getService(self, soapBaseUrl, name):
        sudsCacheDir = os.path.join("%s" %self.cfg['TEMPORARY_DIR'], "suds")
        sudsCache = ObjectCache(sudsCacheDir)

        sudsLogFile = os.path.join(self.cfg['RUNTIME_LOG_DIR'], 'suds.log')
        logging.basicConfig(filename=sudsLogFile)
        logging.getLogger('suds.client').setLevel(self.log.getEffectiveLevel())

        wsdl = "%s/%s?wsdl" % (soapBaseUrl, name)
        return Client(wsdl, cache=sudsCache, location=wsdl)

    def __registerScmIntegration(self, soapBaseUrl, adapterType, name, title, paramNames, paramValues, externalSystemId):

        try:
            sessionId = self.__getService(soapBaseUrl, "CollabNet").service.login("scmadmin", self.cfg["SCM_ADMIN_PASSWORD"])
        except Exception, ex:
            raise Exception("Authenticating with TeamForge (at %s) failed: %r" % (soapBaseUrl, ex))


        scmService = self.__getService(soapBaseUrl, "ScmApp")
        adapterParams = scmService.factory.create('ns1:SoapNamedValues')
        adapterParams.names = paramNames
        adapterParams.values = paramValues

        if externalSystemId != None:
            self.log.info("Updating external sytem %s" % externalSystemId)

            externalSystemSoapDO  = scmService.service.getExternalSystem(sessionId,
                                externalSystemId )

            if externalSystemSoapDO.systemDescription == None and (adapterType == "Subversion" or adapterType == "CVS"):
                externalSystemSoapDO.systemDescription = title

            externalSystemSoapDO.adapterParams = adapterParams
            scmService.service.setExternalSystemData(sessionId,
                                externalSystemSoapDO)
        else:
            exsysid = scmService.service.addExternalSystem(
                sessionId,
                adapterType,
                name,
                title,
                adapterParams)
            self.log.info("Registered %s with ExsyId %s" % (name, exsysid))

    def __isEndpointExternallyReachable(self, url):
        try:
            urllib2.urlopen(url, timeout=2).getcode()
            return True

        except urllib2.HTTPError, e:
            # Any code below 500 means that the server is reachable. A 5xx (503, in particular) could
            # come from a load balancer.
            return e.code < 500

        except Exception, e:
            self.log.warn("URL %s not externally reachable (yet): %r" % (url, e))
            return False

    def __waitForEndpointToBecomeExternallyReachable(self, url):
        # We know that the integration server is up - but if it is running behind a load balancer,
        # it may not be reachable by the application yet (the load balancer may take a few more seconds
        # before it takes the backend into the pool). However, if the integration server is not
        # reachable by the app, validation will fail and the registration will be rejected.
        #
        # Therefore, give the load balancer some slack.
        for i in range(1, 60):
            if self.__isEndpointExternallyReachable(url):
                return
            else:
                self.log.info("Endpoint '%s' not reachable yet, waiting..." % url)
                time.sleep(5)

        raise Exception("URL %s not externally reachable" % url)


    def __registerSvn(self, externalSystemId):
        # Determine own endpoints from the view of the 'ctfcore-app' service (wherever it is deployed)
        coreMachine    = app.landscape.Landscape.getSite().getMachinesByService("ctfcore-app")[0]
        coreAppService = coreMachine.getService("ctfcore-app")

        scheme         = 'https' if self.cfg['SSL'] == 'on' else 'http'

        soapEndpoint   = coreAppService.getRequiredEndpoint("http-soap", "integration").getProviderEndpoint()
        soapServer     = soapEndpoint.getHost(coreMachine.isLocal())
        soapPort       = soapEndpoint.getPort()

        viewvcEndpoint = coreAppService.getRequiredEndpoint("http-web-viewvc", "integration").getProviderEndpoint()
        viewvcServer   = viewvcEndpoint.getHost(coreMachine.isLocal())
        viewvcPort     = viewvcEndpoint.getPort()
        viewvcUrl      = "%s://%s%s/integration/viewvc/viewvc.cgi" % (scheme, viewvcServer, self.__portSegment(scheme, viewvcPort))

        svnEndpoint    = coreAppService.getRequiredEndpoint("http-svn", "integration").getProviderEndpoint()
        svnServer      = svnEndpoint.getHost(coreMachine.isLocal())
        svnPort        = svnEndpoint.getPort()
        svnBaseUrl     = "%s://%s%s%s" % (scheme, svnServer, self.__portSegment(scheme, svnPort), self.cfg['SUBVERSION_URI'])

        ctfSoapUrl     = "%s://%s:%s/ce-soap60/services" % (
                            "http" if self.cfg["integration/dependency/ctfcore-app/http-soap/host"] == "localhost" else scheme,
                            self.cfg["integration/dependency/ctfcore-app/http-soap/host"],
                            self.cfg["integration/dependency/ctfcore-app/http-soap/port"])

        if soapServer != "localhost":
             # Wait for CTF to be reachable for us (Note: By checking the URL, we ensure that not only Apache
            # has opened the port, but that Jboss is actually available.
            self.__waitForEndpointToBecomeExternallyReachable(ctfSoapUrl)

            # Wait for this server to be reachable by CTF.
            self.__waitForEndpointToBecomeExternallyReachable("%s://%s:%d/" % (scheme, soapServer, soapPort))

        paramNames = ['ScmViewerUrl',
             'RepositoryRootPath',
             'RepositoryBaseUrl',
             'HostSSL',
             'HostPort',
             'HostName']
        paramValues = [viewvcUrl,
             self.cfg['SUBVERSION_REPOSITORY_BASE'],
             svnBaseUrl,
             "true" if self.cfg['SSL'] == 'on' and soapServer != "localhost" else "false",
             soapPort,
             soapServer]

        if externalSystemId == None:
            paramNames.extend(['RequireApproval','UseInternalCodeBrowser'])
            paramValues.extend(['false','true' ])

        self.__registerScmIntegration(
            ctfSoapUrl,
            'Subversion',
            "SVN",
            "TeamForge Subversion SCM integration",
            paramNames,
            paramValues,
             externalSystemId)

    def __registerCvs(self, externalSystemId):
        # Determine own endpoints from the view of the 'ctfcore-app' service (wherever it is deployed)
        coreMachine    = app.landscape.Landscape.getSite().getMachinesByService("ctfcore-app")[0]
        coreAppService = coreMachine.getService("ctfcore-app")

        scheme         = 'https' if self.cfg['SSL'] == 'on' else 'http'
        soapEndpoint   = coreAppService.getRequiredEndpoint("http-soap", "integration").getProviderEndpoint()
        soapServer     = soapEndpoint.getHost(coreMachine.isLocal())
        soapPort       = soapEndpoint.getPort()

        viewvcEndpoint = coreAppService.getRequiredEndpoint("http-web-viewvc", "integration").getProviderEndpoint()
        viewvcServer   = viewvcEndpoint.getHost(coreMachine.isLocal())
        viewvcPort     = viewvcEndpoint.getPort()
        viewvcUrl      = "%s://%s%s/integration/viewvc/viewvc.cgi" % (scheme, viewvcServer,  self.__portSegment(scheme, viewvcPort))

        ctfSoapUrl     = "%s://%s:%s/ce-soap60/services" % (
                            "http" if self.cfg["integration/dependency/ctfcore-app/http-soap/host"] == "localhost" else scheme,
                            self.cfg["integration/dependency/ctfcore-app/http-soap/host"],
                            self.cfg["integration/dependency/ctfcore-app/http-soap/port"])

        if soapServer != "localhost":
            # Wait for CTF to be reachable for us (Note: By checking the URL, we ensure that not only Apache
            # has opened the port, but that Jboss is actually available.
            self.__waitForEndpointToBecomeExternallyReachable(ctfSoapUrl)

            # Wait for this server to be reachable by CTF.
            self.__waitForEndpointToBecomeExternallyReachable("%s://%s:%d/" % (scheme, soapServer, soapPort))

        paramNames = ['ScmViewerUrl',
             'RepositoryRootPath',
             'CvsHost',
             'HostSSL',
             'HostPort',
             'HostName' ]

        paramValues = [viewvcUrl,
             self.cfg['CVS_REPOSITORY_BASE'],
             soapServer,
             "true" if self.cfg['SSL'] == 'on' and soapServer != "localhost" else "false",
             soapPort,
             soapServer]

        if externalSystemId == None:
            paramNames.extend(['RequireApproval'])
            paramValues.extend(['false'])

        self.__registerScmIntegration(
            ctfSoapUrl,
            "CVS",
            "CVS",
            "TeamForge CVS SCM integration",
            paramNames,
            paramValues,
             externalSystemId)

    def initialize(self):

        if app.landscape.Landscape.getSite().getLocalMachine().hasService('subversion') and \
           self.cfg.has_key('SUBVERSION_REPOSITORY_BASE') and os.path.isdir(self.cfg['SUBVERSION_REPOSITORY_BASE']):
            self.__registerSvn(self.__getExternalSystemId(self.cfg['SUBVERSION_REPOSITORY_BASE']))

        if app.landscape.Landscape.getSite().getLocalMachine().hasService('cvs') and \
           self.cfg.has_key('CVS_REPOSITORY_BASE') and os.path.isdir(self.cfg['CVS_REPOSITORY_BASE']):
            self.__registerCvs(self.__getExternalSystemId(self.cfg['CVS_REPOSITORY_BASE']))
